<?xml version="1.0" encoding="UTF-8" ?>

<xsl:stylesheet version="3.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform" xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:udf="user-defined" xmlns:chain="chain"
	xmlns:fn="http://www.w3.org/2005/xpath-functions" xmlns:xdt="http://www.w3.org/2005/xpath-datatypes" xmlns:err="http://www.w3.org/2005/xqt-errors"
	xmlns:math="http://exslt.org/math" xmlns:saxon="http://saxon.sf.net/" xmlns:exsl="http://exslt.org/common" xmlns:random="http://exslt.org/random"
	xmlns:date="http://exslt.org/dates-and-times" xmlns:map="http://www.w3.org/2005/xpath-functions/map"
	exclude-result-prefixes="xs xdt udf saxon math fn err exsl date map chain random">
	<xsl:import href="functions.xsl"/>
	
	<xsl:key name="connection.key" match="//target" use="concat(../@x,'--',@y)"/>
	
	<xsl:variable name="max.iterations" select="200"/>
	<xsl:variable name="path" select="replace(base-uri(),'(.*/)[^/]+?.xml','$1')"/>
	
	<!--
	TG1.file and TG2.file are filenames of optional transgraphs used in the processing. These files are not needed and when absent the script
	will bootstrap based on the current graph values.-->
	
	<xsl:variable name="TG1.file" select="concat($path,'Transgraph_Round10_R.xml')"/>
	<xsl:variable name="TG2.file" select="concat($path,'Transgraph_Round10_R.xml')"/>
	<xsl:variable name="transgraph1" select="if (doc-available($TG1.file)) then doc($TG1.file) else ()"/>
	<xsl:variable name="transgraph2" select="if (doc-available($TG2.file)) then doc($TG2.file) else ()"/>

	<xsl:template match="/">
		<xsl:variable name="graph">
			<xsl:copy-of select="graph/from"/>
		</xsl:variable>
		<xsl:result-document href="{concat($path,'DigraphI.xml')}">
	<digraph>
	<xsl:iterate select="1 to $max.iterations">
		<xsl:param name="input" select="$graph"/>
		
		<xsl:if test="round(current() div 10) = current() div 10">
			<xsl:message>
				<a round="{current()}"/>
			</xsl:message>
		</xsl:if>
		
		<!--We determine cases where one edge loses points to other edges. Each edge is considered and 
			one t element i made for each other edge that will take from it. Then a t element is made to represent the overall loss to the
		edge under review.-->

		<xsl:variable name="mid.iteration">
				<!--
				If the input transgraph files exist, we use them, otherwise we use the current digraph itself.
				This allows bootstrapping from a fresh input initial digraph when no transgraphs are known.
				-->
			<xsl:variable name="doc.1" select="($transgraph2, $input)[1]" as="document-node()"/>
			<xsl:variable name="doc.2" select="($transgraph1, $input)[1]" as="document-node()"/>
			
			<xsl:for-each select="$input/from">
				<xsl:variable name="x" select="@x"/>
					<xsl:variable name="targets" select="target[+@c.score gt 0]"/>
				<xsl:copy>
					<xsl:copy-of select="@*"/>
					<xsl:for-each select="target">
						<xsl:variable name="y" select="@y"/>
						
						<xsl:choose>
							<xsl:when test="@edge.score=0">
								<t y="{$y}"/>
							</xsl:when>
							<xsl:otherwise>
						<xsl:variable name="to" select="current()"/>
						
				<!--				For connections with non-zero strength, we look through all the possible intermediary points and see if any
								are strong enough to take some strength away from the from-to pair under consideration.-->
								
								<xsl:variable name="min.score" select="max((0,@edge.score - @red.score))"/>
									
								<xsl:iterate select="$targets[@y ne $y][+@edge.score gt $min.score]">
									<xsl:param name="total.transfer" select="0" as="xs:double"/>
									
							<!--		The co.score is the strength from the intermediary point to the target. This is equal to the average value of the two input transgraphs.-->
									<xsl:variable name="c.string" select="concat(@y,'--',$y)"></xsl:variable>					
									<xsl:variable name="co.score" select="(key('connection.key',$c.string,$doc.2)/@c.score + key('connection.key',$c.string,$doc.1)/@c.score + key('connection.key',$c.string,$input)/@c.score) div 3"/>
									
						<!--			The red.capped value combines all the various redundancy factors. In my implementation I think of the value of R(B,A)
									as a bonus supplied to the other two linkages rather than a penalty to S(A,B). The full value of that bonus is just
									$to/@red.score. The Min() function caps this bonus at a max of 2 times the greater of S(A,M) = @score and S(M,B)=$co.score
									It then subtracts from this the penalty that M,B itself might have because of its own redundancy to A.-->
									
									<xsl:variable name="red.capped" select="min(($to/@red.score, 2 * $co.score)) - @red.score"/>
									
									
									<xsl:choose>
										<xsl:when test="$co.score lt $min.score">
											<xsl:next-iteration/>
										</xsl:when>
										<xsl:when test="($co.score + $red.capped gt +$to/@edge.score) and (@edge.score + $red.capped gt +$to/@edge.score)">
											<xsl:variable name="local.trans" select="0.1 * (@edge.score + $red.capped - $to/@edge.score) * ($co.score  + $red.capped - $to/@edge.score)"/>
											<xsl:variable name="new.total.trans" select="$total.transfer + $local.trans"/>
												<xsl:text>
			</xsl:text>
													<t y="{@y}" d="{-$local.trans}"/>
												
													<xsl:next-iteration>
														<xsl:with-param name="total.transfer" select="$new.total.trans"/>
													</xsl:next-iteration>
										
										</xsl:when>
										<xsl:otherwise>
											<xsl:next-iteration/>
										</xsl:otherwise>
									</xsl:choose>
									<xsl:on-completion>
										
									<!--	N.B. Here we put the total score lost to the edge in question. Note that @y is
										the edge under review rather than the individual candidate edges that robbed points
										from that edge.-->
										
										<xsl:text>
			</xsl:text>
										<t y="{$to/@y}" d="{$total.transfer}"/>
									</xsl:on-completion>
								</xsl:iterate>
							</xsl:otherwise>
						</xsl:choose>
					</xsl:for-each>
				</xsl:copy>
			</xsl:for-each>
		</xsl:variable>
		
<!--		We adjust the edges and create a new graph for the next iteration.
		-->
		<xsl:variable name="next">
			
		<xsl:for-each select="$mid.iteration/from">
	<xsl:text>
	</xsl:text>
			<xsl:copy>
				<xsl:copy-of select="@*"/>
			<xsl:variable name="x" select="@x"/>
			<xsl:for-each-group select="t" group-by="@y">
				<xsl:text>
	</xsl:text>
				<xsl:variable name="y" select="current-grouping-key()"/>
				<xsl:variable name="connection" select="key('connection.key',concat($x,'--',$y),$input)"/>
				<xsl:variable name="transfer" select="min(($connection/@edge.score, sum(current-group()/@d)))"/>
				<target y="{$y}" edge.score="{$connection/@edge.score - $transfer}"
					red.score="{$connection/@red.score}"
					c.score="{$connection/@c.score - $transfer}"
				/>
			</xsl:for-each-group>
			</xsl:copy>
		</xsl:for-each>
		</xsl:variable>
		
		<xsl:next-iteration>
			<xsl:with-param name="input" select="$next"/>
		</xsl:next-iteration>

		<xsl:on-completion>
				<!--	After specified number of passes, we publish the digraph created. We first scale the outputs from
				each node so they add to one. Then we publish the information with extra annotation.-->
				<xsl:variable name="final.scaled">		
					<xsl:for-each select="$next/from">
						<xsl:text>
</xsl:text>
						<xsl:copy>
							<xsl:copy-of select="@*"/>
							<xsl:variable name="targets">
								<xsl:perform-sort select="target">
									<xsl:sort select="-xs:double(@edge.score)"/>
								</xsl:perform-sort>
							</xsl:variable>
							<xsl:variable name="base" select="sum(for $n in 1 to 5 return $targets/target[$n]/@edge.score div $n)"/>
							<xsl:for-each select="$targets/target">
								<xsl:copy>
									<xsl:copy-of select="@*"/>
									<xsl:if test="+@edge.score gt 0">
									<xsl:attribute name="scaled.score" select="math:log($base div @edge.score)"/>
									</xsl:if>
								</xsl:copy>
							</xsl:for-each>
						</xsl:copy>
					</xsl:for-each>
				</xsl:variable>
				
			
					<xsl:for-each select="$final.scaled/from">
						<xsl:variable name="from" select="@x"/>
						<xsl:text>
</xsl:text>
						<xsl:copy>
							<xsl:copy-of select="@*"/>
							<xsl:attribute name="granularity" select="count(target[+@edge.score gt 0])"/>
							<xsl:for-each select="target">
								<xsl:variable name="back.connection" select="key('connection.key',concat(@y,'--',$from),$final.scaled)/@scaled.score"/>
								<xsl:text>
</xsl:text>
								<xsl:copy>
									<xsl:copy-of select="@*"/>
									<xsl:attribute name="back.connection" select="$back.connection"/>
									<xsl:attribute name="connection.type" select="if (+@edge.score gt 0) then
										(if ($back.connection ne '') then 'B' else 'F') else
										(if ($back.connection ne '') then 'R' else 'N')"/>
								</xsl:copy>
							</xsl:for-each>
							
						</xsl:copy>
					</xsl:for-each>
				
		</xsl:on-completion>
	</xsl:iterate>
		</digraph>
		</xsl:result-document>

	</xsl:template>

</xsl:stylesheet>
